from abc import ABCMeta, abstractmethod, abstractproperty

import pyrsistent
import six

from dagster.core.events.log import EventRecord
from dagster.core.execution.stats import (
    build_run_stats_from_events,
    build_run_step_stats_from_events,
)


class EventLogSequence(pyrsistent.CheckedPVector):
    __type__ = EventRecord


class EventLogStorage(six.with_metaclass(ABCMeta)):
    """Abstract base class for storing structured event logs from pipeline runs.

    Note that event log storages using SQL databases as backing stores should implement
    :py:class:`~dagster.core.storage.event_log.SqlEventLogStorage`.

    Users should not directly instantiate concrete subclasses of this class; they are instantiated
    by internal machinery when ``dagit`` and ``dagster-graphql`` load, based on the values in the
    ``dagster.yaml`` file in ``$DAGSTER_HOME``. Configuration of concrete subclasses of this class
    should be done by setting values in that file.
    """

    @abstractmethod
    def get_logs_for_run(self, run_id, cursor=-1):
        """Get all of the logs corresponding to a run.

        Args:
            run_id (str): The id of the run for which to fetch logs.
            cursor (Optional[int]): Zero-indexed logs will be returned starting from cursor + 1,
                i.e., if cursor is -1, all logs will be returned. (default: -1)
        """

    def get_stats_for_run(self, run_id):
        """Get a summary of events that have ocurred in a run."""
        return build_run_stats_from_events(run_id, self.get_logs_for_run(run_id))

    def get_step_stats_for_run(self, run_id):
        """Get per-step stats for a pipeline run."""
        return build_run_step_stats_from_events(run_id, self.get_logs_for_run(run_id))

    @abstractmethod
    def store_event(self, event):
        """Store an event corresponding to a pipeline run.

        Args:
            run_id (str): The id of the run that generated the event.
            event (EventRecord): The event to store.
        """

    @abstractmethod
    def delete_events(self, run_id):
        """Remove events for a given run id"""

    @abstractmethod
    def wipe(self):
        """Clear the log storage."""

    @abstractmethod
    def watch(self, run_id, start_cursor, callback):
        """Call this method to start watching."""

    @abstractmethod
    def end_watch(self, run_id, handler):
        """Call this method to stop watching."""

    @abstractproperty
    def is_persistent(self):
        """bool: Whether the storage is persistent."""

    def dispose(self):
        """Explicit lifecycle management."""

    @property
    def is_asset_aware(self):
        return isinstance(self, AssetAwareEventLogStorage)


class AssetAwareEventLogStorage(six.with_metaclass(ABCMeta)):
    @abstractmethod
    def get_all_asset_keys(self):
        pass

    @abstractmethod
    def get_asset_events(self, asset_key, cursor=None, limit=None):
        pass

    @abstractmethod
    def get_asset_run_ids(self, asset_key):
        pass

    @abstractmethod
    def wipe_asset(self, asset_key):
        """Remove asset index history from event log for given asset_key"""



# encoding: utf-8
import hashlib

from dagster import check
from dagster.core.errors import DagsterInvalidConfigDefinitionError

from .config_type import ConfigType, ConfigTypeKind


def all_optional_type(config_type):
    check.inst_param(config_type, "config_type", ConfigType)

    if ConfigTypeKind.is_shape(config_type.kind):
        for field in config_type.fields.values():
            if field.is_required:
                return False
        return True

    if ConfigTypeKind.is_selector(config_type.kind):
        if len(config_type.fields) == 1:
            for field in config_type.fields.values():
                if field.is_required:
                    return False
            return True

    return False


class __FieldValueSentinel(object):
    pass


class __InferOptionalCompositeFieldSentinel(object):
    pass


FIELD_NO_DEFAULT_PROVIDED = __FieldValueSentinel

INFER_OPTIONAL_COMPOSITE_FIELD = __InferOptionalCompositeFieldSentinel


def check_user_facing_opt_config_param(potential_field, param_name):
    check.str_param(param_name, "param_name")

    return None if potential_field is None else convert_potential_field(potential_field)


class _ConfigHasFields(ConfigType):
    def __init__(self, fields, **kwargs):
        self.fields = expand_fields_dict(fields)
        super(_ConfigHasFields, self).__init__(**kwargs)


FIELD_HASH_CACHE = {}


def _memoize_inst_in_field_cache(passed_cls, defined_cls, key):
    if key in FIELD_HASH_CACHE:
        return FIELD_HASH_CACHE[key]

    defined_cls_inst = super(defined_cls, passed_cls).__new__(defined_cls)

    FIELD_HASH_CACHE[key] = defined_cls_inst
    return defined_cls_inst


def _add_hash(m, string):
    m.update(string.encode())


def _compute_fields_hash(fields, description):

    m = hashlib.sha1()  # so that hexdigest is 40, not 64 bytes
    if description:
        _add_hash(m, ":description: " + description)

    for field_name in sorted(list(fields.keys())):
        field = fields[field_name]
        _add_hash(m, ":fieldname:" + field_name)
        if field.default_provided:
            _add_hash(m, ":default_value: " + field.default_value_as_json_str)
        _add_hash(m, ":is_required: " + str(field.is_required))
        _add_hash(m, ":type_key: " + field.config_type.key)
        if field.description:
            _add_hash(m, ":description: " + field.description)

    return m.hexdigest()


def _define_shape_key_hash(fields, description):
    return "Shape." + _compute_fields_hash(fields, description)


class Shape(_ConfigHasFields):
    """
    Schema for configuration data with string keys and typed values via :py:class:`Field` .

    Args:
        fields (Dict[str, Field])
    """

    def __new__(
        cls, fields, description=None,
    ):
        return _memoize_inst_in_field_cache(
            cls, Shape, _define_shape_key_hash(expand_fields_dict(fields), description),
        )

    def __init__(self, fields, description=None):
        fields = expand_fields_dict(fields)
        super(Shape, self).__init__(
            kind=ConfigTypeKind.STRICT_SHAPE,
            key=_define_shape_key_hash(fields, description),
            description=description,
            fields=fields,
        )


def _define_permissive_dict_key(fields, description):
    return (
        "Permissive." + _compute_fields_hash(fields, description=description)
        if fields
        else "Permissive"
    )


class Permissive(_ConfigHasFields):
    """Defines a config dict with a partially specified schema.

    A permissive dict allows partial specification of the config schema. Any fields with a
    specified schema will be type checked. Other fields will be allowed, but will be ignored by
    the type checker.

    Args:
        fields (Dict[str, Field]): The partial specification of the config dict.

    **Examples**

    .. code-block:: python

        @solid(config_schema=Field(Permissive({'required': Field(String)})))
        def partially_specified_config(context) -> List:
            return sorted(list(context.solid_config.items()))
    """

    def __new__(cls, fields=None, description=None):
        return _memoize_inst_in_field_cache(
            cls,
            Permissive,
            _define_permissive_dict_key(
                expand_fields_dict(fields) if fields else None, description
            ),
        )

    def __init__(self, fields=None, description=None):
        fields = expand_fields_dict(fields) if fields else None
        super(Permissive, self).__init__(
            key=_define_permissive_dict_key(fields, description),
            kind=ConfigTypeKind.PERMISSIVE_SHAPE,
            fields=fields or dict(),
            description=description,
        )


def _define_selector_key(fields, description):
    return "Selector." + _compute_fields_hash(fields, description=description)


class Selector(_ConfigHasFields):
    """Define a config field requiring the user to select one option.

    Selectors are used when you want to be able to present several different options in config but
    allow only one to be selected. For example, a single input might be read in from either a csv
    file or a parquet file, but not both at once.

    Note that in some other type systems this might be called an 'input union'.

    Functionally, a selector is like a :py:class:`Dict`, except that only one key from the dict can
    be specified in valid config.

    Args:
        fields (Dict[str, Field]): The fields from which the user must select.

    Examples:

    .. code-block:: python

        @solid(
            config_schema=Field(
                Selector(
                    {
                        'haw': {'whom': Field(String, default_value='honua', is_required=False)},
                        'cn': {'whom': Field(String, default_value='世界', is_required=False)},
                        'en': {'whom': Field(String, default_value='world', is_required=False)},
                    }
                ),
                is_required=False,
                default_value={'en': {'whom': 'world'}},
            )
        )
        def hello_world_with_default(context):
            if 'haw' in context.solid_config:
                return 'Aloha {whom}!'.format(whom=context.solid_config['haw']['whom'])
            if 'cn' in context.solid_config:
                return '你好，{whom}!'.format(whom=context.solid_config['cn']['whom'])
            if 'en' in context.solid_config:
                return 'Hello, {whom}!'.format(whom=context.solid_config['en']['whom'])
    """

    def __new__(cls, fields, description=None):
        return _memoize_inst_in_field_cache(
            cls, Selector, _define_selector_key(expand_fields_dict(fields), description),
        )

    def __init__(self, fields, description=None):
        fields = expand_fields_dict(fields)
        super(Selector, self).__init__(
            key=_define_selector_key(fields, description),
            kind=ConfigTypeKind.SELECTOR,
            fields=fields,
            description=description,
        )


# Config syntax expansion code below


def is_potential_field(potential_field):
    from .field import Field, resolve_to_config_type

    return isinstance(potential_field, (Field, dict, list)) or resolve_to_config_type(
        potential_field
    )


def convert_fields_to_dict_type(fields):
    return _convert_fields_to_dict_type(fields, fields, [])


def _convert_fields_to_dict_type(original_root, fields, stack):
    return Shape(_expand_fields_dict(original_root, fields, stack))


def expand_fields_dict(fields):
    return _expand_fields_dict(fields, fields, [])


def _expand_fields_dict(original_root, fields, stack):
    check.dict_param(fields, "fields")
    return {
        name: _convert_potential_field(original_root, value, stack + [name])
        for name, value in fields.items()
    }


def expand_list(original_root, the_list, stack):
    from .config_type import Array

    if len(the_list) != 1:
        raise DagsterInvalidConfigDefinitionError(
            original_root, the_list, stack, "List must be of length 1"
        )

    inner_type = _convert_potential_type(original_root, the_list[0], stack)
    if not inner_type:
        raise DagsterInvalidConfigDefinitionError(
            original_root,
            the_list,
            stack,
            "List have a single item and contain a valid type i.e. [int]. Got item {}".format(
                repr(the_list[0])
            ),
        )

    return Array(inner_type)


def convert_potential_field(potential_field):
    return _convert_potential_field(potential_field, potential_field, [])


def _convert_potential_type(original_root, potential_type, stack):
    from .field import resolve_to_config_type

    if isinstance(potential_type, dict):
        return Shape(_expand_fields_dict(original_root, potential_type, stack))

    if isinstance(potential_type, list):
        return expand_list(original_root, potential_type, stack)

    return resolve_to_config_type(potential_type)


def _convert_potential_field(original_root, potential_field, stack):
    from .field import Field

    if potential_field is None:
        raise DagsterInvalidConfigDefinitionError(
            original_root, potential_field, stack, reason="Fields cannot be None"
        )

    if not is_potential_field(potential_field):
        raise DagsterInvalidConfigDefinitionError(original_root, potential_field, stack)

    if isinstance(potential_field, Field):
        return potential_field

    return Field(_convert_potential_type(original_root, potential_field, stack))



import random
import string
import uuid
import warnings

import toposort as toposort_

from dagster.version import __version__

BACKFILL_TAG_LENGTH = 8


def toposort(data):
    return [sorted(list(level)) for level in toposort_.toposort(data)]


def toposort_flatten(data):
    return [item for level in toposort(data) for item in level]


def make_new_run_id():
    return str(uuid.uuid4())


def make_new_backfill_id():
    return "".join(random.choice(string.ascii_lowercase) for x in range(BACKFILL_TAG_LENGTH))


def str_format_list(items):
    return "[{items}]".format(items=", ".join(["'{item}'".format(item=item) for item in items]))


def str_format_set(items):
    return "[{items}]".format(items=", ".join(["'{item}'".format(item=item) for item in items]))


def check_dagster_package_version(library_name, library_version):
    if __version__ != library_version:
        message = "Found version mismatch between `dagster` ({}) and `{}` ({})".format(
            __version__, library_name, library_version
        )
        warnings.warn(message)



import time

from dagster import check
from dagster.core.events import DagsterEvent
from dagster.core.execution.retries import Retries

from .plan import ExecutionPlan


def _default_sort_key(step):
    return int(step.tags.get("dagster/priority", 0)) * -1


class ActiveExecution(object):
    """State machine used to track progress through execution of an ExecutionPlan
    """

    def __init__(self, execution_plan, retries, sort_key_fn=None):
        self._plan = check.inst_param(execution_plan, "execution_plan", ExecutionPlan)
        self._retries = check.inst_param(retries, "retries", Retries)
        self._sort_key_fn = check.opt_callable_param(sort_key_fn, "sort_key_fn", _default_sort_key)

        # All steps to be executed start out here in _pending
        self._pending = self._plan.execution_deps()

        # steps move in to these buckets as a result of _update calls
        self._executable = []
        self._pending_skip = []
        self._pending_retry = []
        self._waiting_to_retry = {}

        # then are considered _in_flight when vended via get_steps_to_*
        self._in_flight = set()

        # and finally their terminal state is tracked by these sets, via mark_*
        self._completed = set()
        self._success = set()
        self._failed = set()
        self._skipped = set()

        # Start the show by loading _executable with the set of _pending steps that have no deps
        self._update()

    def _update(self):
        """Moves steps from _pending to _executable / _pending_skip / _pending_retry
           as a function of what has been _completed
        """
        new_steps_to_execute = []
        new_steps_to_skip = []
        for step_key, requirements in self._pending.items():
            if requirements.issubset(self._completed):
                if requirements.issubset(self._success):
                    new_steps_to_execute.append(step_key)
                else:
                    step = self.get_step_by_key(step_key)

                    # The step will skip on any upstream failure
                    should_skip = True

                    # Unless a fan-in input has any successful inputs
                    for inp in step.step_inputs:
                        if inp.is_from_multiple_outputs:
                            if any([key in self._success for key in inp.dependency_keys]):
                                should_skip = False

                    # but no missing regular inputs
                    for inp in step.step_inputs:
                        if inp.is_from_single_output:
                            if any([key not in self._success for key in inp.dependency_keys]):
                                should_skip = True

                    if should_skip:
                        new_steps_to_skip.append(step_key)
                    else:
                        new_steps_to_execute.append(step_key)

        for key in new_steps_to_execute:
            self._executable.append(key)
            del self._pending[key]

        for key in new_steps_to_skip:
            self._pending_skip.append(key)
            del self._pending[key]

        ready_to_retry = []
        tick_time = time.time()
        for key, at_time in self._waiting_to_retry.items():
            if tick_time >= at_time:
                ready_to_retry.append(key)

        for key in ready_to_retry:
            self._executable.append(key)
            del self._waiting_to_retry[key]

    def sleep_til_ready(self):
        now = time.time()
        sleep_amt = min([ready_at - now for ready_at in self._waiting_to_retry.values()])
        if sleep_amt > 0:
            time.sleep(sleep_amt)

    def get_next_step(self):
        check.invariant(not self.is_complete, "Can not call get_next_step when is_complete is True")

        steps = self.get_steps_to_execute(limit=1)
        step = None

        if steps:
            step = steps[0]
        elif self._waiting_to_retry:
            self.sleep_til_ready()
            step = self.get_next_step()

        check.invariant(step is not None, "Unexpected ActiveExecution state")
        return step

    def get_step_by_key(self, step_key):
        return self._plan.get_step_by_key(step_key)

    def get_steps_to_execute(self, limit=None):
        check.opt_int_param(limit, "limit")
        self._update()

        steps = sorted(
            [self.get_step_by_key(key) for key in self._executable], key=self._sort_key_fn
        )

        if limit:
            steps = steps[:limit]

        for step in steps:
            self._in_flight.add(step.key)
            self._executable.remove(step.key)
        return steps

    def get_steps_to_skip(self):
        self._update()

        steps = []
        steps_to_skip = list(self._pending_skip)
        for key in steps_to_skip:
            steps.append(self.get_step_by_key(key))
            self._in_flight.add(key)
            self._pending_skip.remove(key)

        return sorted(steps, key=self._sort_key_fn)

    def skipped_step_events_iterator(self, pipeline_context):
        """Process all steps that can be skipped by repeated calls to get_steps_to_skip
        """
        failed_or_skipped_steps = self._skipped.union(self._failed)

        steps_to_skip = self.get_steps_to_skip()
        while steps_to_skip:
            for step in steps_to_skip:
                step_context = pipeline_context.for_step(step)
                failed_inputs = []
                for step_input in step.step_inputs:
                    failed_inputs.extend(
                        failed_or_skipped_steps.intersection(step_input.dependency_keys)
                    )

                step_context.log.info(
                    "Dependencies for step {step} failed: {failed_inputs}. Not executing.".format(
                        step=step.key, failed_inputs=failed_inputs
                    )
                )
                yield DagsterEvent.step_skipped_event(step_context)

                self.mark_skipped(step.key)

            steps_to_skip = self.get_steps_to_skip()

    def mark_failed(self, step_key):
        self._failed.add(step_key)
        self._mark_complete(step_key)

    def mark_success(self, step_key):
        self._success.add(step_key)
        self._mark_complete(step_key)

    def mark_skipped(self, step_key):
        self._skipped.add(step_key)
        self._mark_complete(step_key)

    def mark_up_for_retry(self, step_key, at_time=None):
        check.invariant(
            not self._retries.disabled,
            "Attempted to mark {} as up for retry but retries are disabled".format(step_key),
        )
        check.opt_float_param(at_time, "at_time")

        # if retries are enabled - queue this back up
        if self._retries.enabled:
            if at_time:
                self._waiting_to_retry[step_key] = at_time
            else:
                self._pending[step_key] = self._plan.execution_deps()[step_key]

        elif self._retries.deferred:
            self._completed.add(step_key)

        self._retries.mark_attempt(step_key)
        self._in_flight.remove(step_key)

    def _mark_complete(self, step_key):
        check.invariant(
            step_key not in self._completed,
            "Attempted to mark step {} as complete that was already completed".format(step_key),
        )
        check.invariant(
            step_key in self._in_flight,
            "Attempted to mark step {} as complete that was not known to be in flight".format(
                step_key
            ),
        )
        self._in_flight.remove(step_key)
        self._completed.add(step_key)

    def handle_event(self, dagster_event):
        check.inst_param(dagster_event, "dagster_event", DagsterEvent)

        if dagster_event.is_step_failure:
            self.mark_failed(dagster_event.step_key)
        elif dagster_event.is_step_success:
            self.mark_success(dagster_event.step_key)
        elif dagster_event.is_step_skipped:
            self.mark_skipped(dagster_event.step_key)
        elif dagster_event.is_step_up_for_retry:
            self.mark_up_for_retry(
                dagster_event.step_key,
                time.time() + dagster_event.step_retry_data.seconds_to_wait
                if dagster_event.step_retry_data.seconds_to_wait
                else None,
            )

    def verify_complete(self, pipeline_context, step_key):
        """Ensure that a step has reached a terminal state, if it has not mark it as an unexpected failure
        """
        if step_key in self._in_flight:
            pipeline_context.log.error(
                "Step {key} finished without success or failure event, assuming failure.".format(
                    key=step_key
                )
            )
            self.mark_failed(step_key)

    @property
    def is_complete(self):
        return (
            len(self._pending) == 0
            and len(self._in_flight) == 0
            and len(self._executable) == 0
            and len(self._pending_skip) == 0
            and len(self._pending_retry) == 0
            and len(self._waiting_to_retry) == 0
        )



# @generated

# This file was generated by running `python -m dagster.grpc.compile`
# Do not edit this file directly, and do not attempt to recompile it using
# grpc_tools.protoc directly, as several changes must be made to the raw output

# pylint: disable=no-member, unused-argument

# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import api_pb2 as api__pb2


class DagsterApiStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Ping = channel.unary_unary(
            "/api.DagsterApi/Ping",
            request_serializer=api__pb2.PingRequest.SerializeToString,
            response_deserializer=api__pb2.PingReply.FromString,
        )
        self.Heartbeat = channel.unary_unary(
            "/api.DagsterApi/Heartbeat",
            request_serializer=api__pb2.PingRequest.SerializeToString,
            response_deserializer=api__pb2.PingReply.FromString,
        )
        self.StreamingPing = channel.unary_stream(
            "/api.DagsterApi/StreamingPing",
            request_serializer=api__pb2.StreamingPingRequest.SerializeToString,
            response_deserializer=api__pb2.StreamingPingEvent.FromString,
        )
        self.ExecutionPlanSnapshot = channel.unary_unary(
            "/api.DagsterApi/ExecutionPlanSnapshot",
            request_serializer=api__pb2.ExecutionPlanSnapshotRequest.SerializeToString,
            response_deserializer=api__pb2.ExecutionPlanSnapshotReply.FromString,
        )
        self.ListRepositories = channel.unary_unary(
            "/api.DagsterApi/ListRepositories",
            request_serializer=api__pb2.ListRepositoriesRequest.SerializeToString,
            response_deserializer=api__pb2.ListRepositoriesReply.FromString,
        )
        self.ExternalPartitionNames = channel.unary_unary(
            "/api.DagsterApi/ExternalPartitionNames",
            request_serializer=api__pb2.ExternalPartitionNamesRequest.SerializeToString,
            response_deserializer=api__pb2.ExternalPartitionNamesReply.FromString,
        )
        self.ExternalPartitionConfig = channel.unary_unary(
            "/api.DagsterApi/ExternalPartitionConfig",
            request_serializer=api__pb2.ExternalPartitionConfigRequest.SerializeToString,
            response_deserializer=api__pb2.ExternalPartitionConfigReply.FromString,
        )
        self.ExternalPartitionTags = channel.unary_unary(
            "/api.DagsterApi/ExternalPartitionTags",
            request_serializer=api__pb2.ExternalPartitionTagsRequest.SerializeToString,
            response_deserializer=api__pb2.ExternalPartitionTagsReply.FromString,
        )
        self.ExternalPartitionSetExecutionParams = channel.unary_unary(
            "/api.DagsterApi/ExternalPartitionSetExecutionParams",
            request_serializer=api__pb2.ExternalPartitionSetExecutionParamsRequest.SerializeToString,
            response_deserializer=api__pb2.ExternalPartitionSetExecutionParamsReply.FromString,
        )
        self.ExternalPipelineSubsetSnapshot = channel.unary_unary(
            "/api.DagsterApi/ExternalPipelineSubsetSnapshot",
            request_serializer=api__pb2.ExternalPipelineSubsetSnapshotRequest.SerializeToString,
            response_deserializer=api__pb2.ExternalPipelineSubsetSnapshotReply.FromString,
        )
        self.ExternalRepository = channel.unary_unary(
            "/api.DagsterApi/ExternalRepository",
            request_serializer=api__pb2.ExternalRepositoryRequest.SerializeToString,
            response_deserializer=api__pb2.ExternalRepositoryReply.FromString,
        )
        self.ExternalScheduleExecution = channel.unary_unary(
            "/api.DagsterApi/ExternalScheduleExecution",
            request_serializer=api__pb2.ExternalScheduleExecutionRequest.SerializeToString,
            response_deserializer=api__pb2.ExternalScheduleExecutionReply.FromString,
        )
        self.ExternalTriggerExecutionParams = channel.unary_unary(
            "/api.DagsterApi/ExternalTriggerExecutionParams",
            request_serializer=api__pb2.ExternalTriggerExecutionParamsRequest.SerializeToString,
            response_deserializer=api__pb2.ExternalTriggerExecutionParamsReply.FromString,
        )
        self.ShutdownServer = channel.unary_unary(
            "/api.DagsterApi/ShutdownServer",
            request_serializer=api__pb2.Empty.SerializeToString,
            response_deserializer=api__pb2.ShutdownServerReply.FromString,
        )
        self.ExecuteRun = channel.unary_stream(
            "/api.DagsterApi/ExecuteRun",
            request_serializer=api__pb2.ExecuteRunRequest.SerializeToString,
            response_deserializer=api__pb2.ExecuteRunEvent.FromString,
        )
        self.CancelExecution = channel.unary_unary(
            "/api.DagsterApi/CancelExecution",
            request_serializer=api__pb2.CancelExecutionRequest.SerializeToString,
            response_deserializer=api__pb2.CancelExecutionReply.FromString,
        )
        self.CanCancelExecution = channel.unary_unary(
            "/api.DagsterApi/CanCancelExecution",
            request_serializer=api__pb2.CanCancelExecutionRequest.SerializeToString,
            response_deserializer=api__pb2.CanCancelExecutionReply.FromString,
        )
        self.StartRun = channel.unary_unary(
            "/api.DagsterApi/StartRun",
            request_serializer=api__pb2.StartRunRequest.SerializeToString,
            response_deserializer=api__pb2.StartRunReply.FromString,
        )
        self.GetCurrentImage = channel.unary_unary(
            "/api.DagsterApi/GetCurrentImage",
            request_serializer=api__pb2.Empty.SerializeToString,
            response_deserializer=api__pb2.GetCurrentImageReply.FromString,
        )


class DagsterApiServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Ping(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def Heartbeat(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def StreamingPing(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ExecutionPlanSnapshot(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ListRepositories(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ExternalPartitionNames(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ExternalPartitionConfig(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ExternalPartitionTags(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ExternalPartitionSetExecutionParams(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ExternalPipelineSubsetSnapshot(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ExternalRepository(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ExternalScheduleExecution(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ExternalTriggerExecutionParams(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ShutdownServer(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ExecuteRun(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def CancelExecution(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def CanCancelExecution(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def StartRun(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetCurrentImage(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_DagsterApiServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "Ping": grpc.unary_unary_rpc_method_handler(
            servicer.Ping,
            request_deserializer=api__pb2.PingRequest.FromString,
            response_serializer=api__pb2.PingReply.SerializeToString,
        ),
        "Heartbeat": grpc.unary_unary_rpc_method_handler(
            servicer.Heartbeat,
            request_deserializer=api__pb2.PingRequest.FromString,
            response_serializer=api__pb2.PingReply.SerializeToString,
        ),
        "StreamingPing": grpc.unary_stream_rpc_method_handler(
            servicer.StreamingPing,
            request_deserializer=api__pb2.StreamingPingRequest.FromString,
            response_serializer=api__pb2.StreamingPingEvent.SerializeToString,
        ),
        "ExecutionPlanSnapshot": grpc.unary_unary_rpc_method_handler(
            servicer.ExecutionPlanSnapshot,
            request_deserializer=api__pb2.ExecutionPlanSnapshotRequest.FromString,
            response_serializer=api__pb2.ExecutionPlanSnapshotReply.SerializeToString,
        ),
        "ListRepositories": grpc.unary_unary_rpc_method_handler(
            servicer.ListRepositories,
            request_deserializer=api__pb2.ListRepositoriesRequest.FromString,
            response_serializer=api__pb2.ListRepositoriesReply.SerializeToString,
        ),
        "ExternalPartitionNames": grpc.unary_unary_rpc_method_handler(
            servicer.ExternalPartitionNames,
            request_deserializer=api__pb2.ExternalPartitionNamesRequest.FromString,
            response_serializer=api__pb2.ExternalPartitionNamesReply.SerializeToString,
        ),
        "ExternalPartitionConfig": grpc.unary_unary_rpc_method_handler(
            servicer.ExternalPartitionConfig,
            request_deserializer=api__pb2.ExternalPartitionConfigRequest.FromString,
            response_serializer=api__pb2.ExternalPartitionConfigReply.SerializeToString,
        ),
        "ExternalPartitionTags": grpc.unary_unary_rpc_method_handler(
            servicer.ExternalPartitionTags,
            request_deserializer=api__pb2.ExternalPartitionTagsRequest.FromString,
            response_serializer=api__pb2.ExternalPartitionTagsReply.SerializeToString,
        ),
        "ExternalPartitionSetExecutionParams": grpc.unary_unary_rpc_method_handler(
            servicer.ExternalPartitionSetExecutionParams,
            request_deserializer=api__pb2.ExternalPartitionSetExecutionParamsRequest.FromString,
            response_serializer=api__pb2.ExternalPartitionSetExecutionParamsReply.SerializeToString,
        ),
        "ExternalPipelineSubsetSnapshot": grpc.unary_unary_rpc_method_handler(
            servicer.ExternalPipelineSubsetSnapshot,
            request_deserializer=api__pb2.ExternalPipelineSubsetSnapshotRequest.FromString,
            response_serializer=api__pb2.ExternalPipelineSubsetSnapshotReply.SerializeToString,
        ),
        "ExternalRepository": grpc.unary_unary_rpc_method_handler(
            servicer.ExternalRepository,
            request_deserializer=api__pb2.ExternalRepositoryRequest.FromString,
            response_serializer=api__pb2.ExternalRepositoryReply.SerializeToString,
        ),
        "ExternalScheduleExecution": grpc.unary_unary_rpc_method_handler(
            servicer.ExternalScheduleExecution,
            request_deserializer=api__pb2.ExternalScheduleExecutionRequest.FromString,
            response_serializer=api__pb2.ExternalScheduleExecutionReply.SerializeToString,
        ),
        "ExternalTriggerExecutionParams": grpc.unary_unary_rpc_method_handler(
            servicer.ExternalTriggerExecutionParams,
            request_deserializer=api__pb2.ExternalTriggerExecutionParamsRequest.FromString,
            response_serializer=api__pb2.ExternalTriggerExecutionParamsReply.SerializeToString,
        ),
        "ShutdownServer": grpc.unary_unary_rpc_method_handler(
            servicer.ShutdownServer,
            request_deserializer=api__pb2.Empty.FromString,
            response_serializer=api__pb2.ShutdownServerReply.SerializeToString,
        ),
        "ExecuteRun": grpc.unary_stream_rpc_method_handler(
            servicer.ExecuteRun,
            request_deserializer=api__pb2.ExecuteRunRequest.FromString,
            response_serializer=api__pb2.ExecuteRunEvent.SerializeToString,
        ),
        "CancelExecution": grpc.unary_unary_rpc_method_handler(
            servicer.CancelExecution,
            request_deserializer=api__pb2.CancelExecutionRequest.FromString,
            response_serializer=api__pb2.CancelExecutionReply.SerializeToString,
        ),
        "CanCancelExecution": grpc.unary_unary_rpc_method_handler(
            servicer.CanCancelExecution,
            request_deserializer=api__pb2.CanCancelExecutionRequest.FromString,
            response_serializer=api__pb2.CanCancelExecutionReply.SerializeToString,
        ),
        "StartRun": grpc.unary_unary_rpc_method_handler(
            servicer.StartRun,
            request_deserializer=api__pb2.StartRunRequest.FromString,
            response_serializer=api__pb2.StartRunReply.SerializeToString,
        ),
        "GetCurrentImage": grpc.unary_unary_rpc_method_handler(
            servicer.GetCurrentImage,
            request_deserializer=api__pb2.Empty.FromString,
            response_serializer=api__pb2.GetCurrentImageReply.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler("api.DagsterApi", rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class DagsterApi(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Ping(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/api.DagsterApi/Ping",
            api__pb2.PingRequest.SerializeToString,
            api__pb2.PingReply.FromString,
            options,
            channel_credentials,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def Heartbeat(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/api.DagsterApi/Heartbeat",
            api__pb2.PingRequest.SerializeToString,
            api__pb2.PingReply.FromString,
            options,
            channel_credentials,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def StreamingPing(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_stream(
            request,
            target,
            "/api.DagsterApi/StreamingPing",
            api__pb2.StreamingPingRequest.SerializeToString,
            api__pb2.StreamingPingEvent.FromString,
            options,
            channel_credentials,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ExecutionPlanSnapshot(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/api.DagsterApi/ExecutionPlanSnapshot",
            api__pb2.ExecutionPlanSnapshotRequest.SerializeToString,
            api__pb2.ExecutionPlanSnapshotReply.FromString,
            options,
            channel_credentials,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ListRepositories(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/api.DagsterApi/ListRepositories",
            api__pb2.ListRepositoriesRequest.SerializeToString,
            api__pb2.ListRepositoriesReply.FromString,
            options,
            channel_credentials,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ExternalPartitionNames(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/api.DagsterApi/ExternalPartitionNames",
            api__pb2.ExternalPartitionNamesRequest.SerializeToString,
            api__pb2.ExternalPartitionNamesReply.FromString,
            options,
            channel_credentials,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ExternalPartitionConfig(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/api.DagsterApi/ExternalPartitionConfig",
            api__pb2.ExternalPartitionConfigRequest.SerializeToString,
            api__pb2.ExternalPartitionConfigReply.FromString,
            options,
            channel_credentials,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ExternalPartitionTags(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/api.DagsterApi/ExternalPartitionTags",
            api__pb2.ExternalPartitionTagsRequest.SerializeToString,
            api__pb2.ExternalPartitionTagsReply.FromString,
            options,
            channel_credentials,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ExternalPartitionSetExecutionParams(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/api.DagsterApi/ExternalPartitionSetExecutionParams",
            api__pb2.ExternalPartitionSetExecutionParamsRequest.SerializeToString,
            api__pb2.ExternalPartitionSetExecutionParamsReply.FromString,
            options,
            channel_credentials,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ExternalPipelineSubsetSnapshot(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/api.DagsterApi/ExternalPipelineSubsetSnapshot",
            api__pb2.ExternalPipelineSubsetSnapshotRequest.SerializeToString,
            api__pb2.ExternalPipelineSubsetSnapshotReply.FromString,
            options,
            channel_credentials,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ExternalRepository(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/api.DagsterApi/ExternalRepository",
            api__pb2.ExternalRepositoryRequest.SerializeToString,
            api__pb2.ExternalRepositoryReply.FromString,
            options,
            channel_credentials,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ExternalScheduleExecution(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/api.DagsterApi/ExternalScheduleExecution",
            api__pb2.ExternalScheduleExecutionRequest.SerializeToString,
            api__pb2.ExternalScheduleExecutionReply.FromString,
            options,
            channel_credentials,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ExternalTriggerExecutionParams(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/api.DagsterApi/ExternalTriggerExecutionParams",
            api__pb2.ExternalTriggerExecutionParamsRequest.SerializeToString,
            api__pb2.ExternalTriggerExecutionParamsReply.FromString,
            options,
            channel_credentials,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ShutdownServer(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/api.DagsterApi/ShutdownServer",
            api__pb2.Empty.SerializeToString,
            api__pb2.ShutdownServerReply.FromString,
            options,
            channel_credentials,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ExecuteRun(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_stream(
            request,
            target,
            "/api.DagsterApi/ExecuteRun",
            api__pb2.ExecuteRunRequest.SerializeToString,
            api__pb2.ExecuteRunEvent.FromString,
            options,
            channel_credentials,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def CancelExecution(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/api.DagsterApi/CancelExecution",
            api__pb2.CancelExecutionRequest.SerializeToString,
            api__pb2.CancelExecutionReply.FromString,
            options,
            channel_credentials,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def CanCancelExecution(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/api.DagsterApi/CanCancelExecution",
            api__pb2.CanCancelExecutionRequest.SerializeToString,
            api__pb2.CanCancelExecutionReply.FromString,
            options,
            channel_credentials,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def StartRun(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/api.DagsterApi/StartRun",
            api__pb2.StartRunRequest.SerializeToString,
            api__pb2.StartRunReply.FromString,
            options,
            channel_credentials,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetCurrentImage(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/api.DagsterApi/GetCurrentImage",
            api__pb2.Empty.SerializeToString,
            api__pb2.GetCurrentImageReply.FromString,
            options,
            channel_credentials,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )



"""
This file contains a set of utilities for dealing with test
environments where we want to go back and forth between
abstractions that reside in user process (e.g. definitions and
reconstructables) and abstractions that reside in host processes
(e.g. handles and externals).

These should only be invoked from contexts where we know this
to be the case.
"""

import sys

from dagster import check
from dagster.core.code_pointer import CodePointer
from dagster.core.definitions.reconstructable import (
    ReconstructableRepository,
    repository_def_from_pointer,
)
from dagster.core.host_representation import ExternalPipeline, ExternalRepository, RepositoryHandle
from dagster.core.host_representation.external_data import (
    external_pipeline_data_from_def,
    external_repository_data_from_def,
)
from dagster.core.host_representation.handle import (
    InProcessRepositoryLocationHandle,
    PythonEnvRepositoryLocationHandle,
    RepositoryLocationHandle,
)
from dagster.core.origin import PipelinePythonOrigin, RepositoryOrigin


def recon_pipeline_from_origin(origin):
    check.inst_param(origin, "origin", PipelinePythonOrigin)
    recon_repo = ReconstructableRepository(origin.get_repo_pointer())
    return recon_repo.get_reconstructable_pipeline(origin.pipeline_name)


def recon_repository_from_origin(origin):
    check.inst_param(origin, "origin", RepositoryOrigin)
    return ReconstructableRepository(origin.code_pointer)


def external_repo_from_def(repository_def, repository_handle):
    return ExternalRepository(external_repository_data_from_def(repository_def), repository_handle)


def recon_repo_from_external_repo(external_repo):
    return ReconstructableRepository(external_repo.get_origin().code_pointer)


def external_pipeline_from_recon_pipeline(recon_pipeline, solid_selection, repository_handle):
    if solid_selection:
        sub_pipeline = recon_pipeline.subset_for_execution(solid_selection)
        pipeline_def = sub_pipeline.get_definition()
    else:
        pipeline_def = recon_pipeline.get_definition()

    return ExternalPipeline(
        external_pipeline_data_from_def(pipeline_def), repository_handle=repository_handle,
    )


def is_repository_location_in_same_python_env(repository_location_handle):
    # either this directly in-process
    if isinstance(repository_location_handle, InProcessRepositoryLocationHandle):
        return True

    # or it is out-of-process but using the same python executable
    return (
        isinstance(repository_location_handle, PythonEnvRepositoryLocationHandle)
        and repository_location_handle.executable_path == sys.executable
    )


def repository_def_from_repository_handle(repository_handle):
    check.inst_param(repository_handle, "repository_handle", RepositoryHandle)
    check.param_invariant(
        is_repository_location_in_same_python_env(repository_handle.repository_location_handle),
        "repository_handle",
        "In order to use this function the location of the repository must be in process "
        "or it must a python environment with the exact same executable.",
    )
    return repository_def_from_pointer(repository_handle.get_origin().code_pointer)


def create_in_process_ephemeral_workspace(pointer):
    from dagster.cli.workspace.workspace import Workspace

    check.inst_param(pointer, "pointer", CodePointer)
    repository_location_handles = [RepositoryLocationHandle.create_in_process_location(pointer)]
    workspace = Workspace(repository_location_handles=repository_location_handles)
    return workspace



import sqlalchemy as db

ScheduleStorageSqlMetadata = db.MetaData()

ScheduleTable = db.Table(
    "schedules",
    ScheduleStorageSqlMetadata,
    db.Column("id", db.Integer, primary_key=True, autoincrement=True),
    db.Column("schedule_origin_id", db.String(255), unique=True),
    db.Column("repository_origin_id", db.String(255)),
    db.Column("status", db.String(63)),
    db.Column("schedule_body", db.String),
    db.Column("create_timestamp", db.DateTime, server_default=db.text("CURRENT_TIMESTAMP")),
    db.Column("update_timestamp", db.DateTime, server_default=db.text("CURRENT_TIMESTAMP")),
)

ScheduleTickTable = db.Table(
    "schedule_ticks",
    ScheduleStorageSqlMetadata,
    db.Column("id", db.Integer, primary_key=True, autoincrement=True),
    db.Column("schedule_origin_id", db.String(255), index=True),
    db.Column("status", db.String(63)),
    db.Column("timestamp", db.types.TIMESTAMP),
    db.Column("tick_body", db.String),
    # The create and update timestamps are not used in framework code, are are simply
    # present for debugging purposes.
    db.Column("create_timestamp", db.DateTime, server_default=db.text("CURRENT_TIMESTAMP")),
    db.Column("update_timestamp", db.DateTime, server_default=db.text("CURRENT_TIMESTAMP")),
)



import hashlib

from dagster import check


def _resolve_step_input_versions(step, step_versions):
    def _resolve_output_version(step_output_handle):
        if (
            step_output_handle.step_key not in step_versions
            or not step_versions[step_output_handle.step_key]
        ):
            return None
        else:
            return join_and_hash(
                [step_versions[step_output_handle.step_key], step_output_handle.output_name]
            )

    input_versions = {}
    for input_name, step_input in step.step_input_dict.items():
        if step_input.is_from_output:
            output_handle_versions = [
                _resolve_output_version(source_handle)
                for source_handle in step_input.source_handles
            ]
            version = join_and_hash(output_handle_versions)
        elif step_input.is_from_config:
            version = step_input.dagster_type.loader.compute_loaded_input_version(
                step_input.config_data
            )
        input_versions[input_name] = version

    return input_versions


def resolve_step_versions(speculative_execution_plan):
    """Resolves the version of each step in an execution plan.

    If an execution plan is not provided, then it is constructed from pipeline_def, run_config, and
    mode. It returns dict[str, str] where each key is a step key, and each value is the associated
    version for that step.

    Args:
        speculative_execution_plan (ExecutionPlan): Execution plan to resolve steps for.
        mode (Optional[str]): The pipeline mode in which to execute this run.
    """
    from dagster.core.execution.plan.plan import ExecutionPlan

    check.inst_param(speculative_execution_plan, "speculative_execution_plan", ExecutionPlan)
    step_versions = {}  # step_key (str) -> version (str)

    for step in speculative_execution_plan.topological_steps():
        input_version_dict = _resolve_step_input_versions(step, step_versions)
        input_versions = [version for version in input_version_dict.values()]

        solid_version = step.solid_version

        from_versions = input_versions + [solid_version]

        step_version = join_and_hash(from_versions)

        step_versions[step.key] = step_version

    return step_versions


def join_and_hash(lst):
    lst = check.list_param(lst, "lst")
    lst = [check.opt_str_param(elem, "elem") for elem in lst]
    if None in lst:
        return None
    else:
        unhashed = "".join(sorted(lst))
        return hashlib.sha1(unhashed.encode("utf-8")).hexdigest()


def resolve_config_version(config_value):
    """Resolve a configuration value into a hashed version.

    If a single value is passed in, it is converted to a string, hashed, and returned as the
    version. If a dictionary of config values is passed in, each value is resolved to a version,
    concatenated with its key, joined, and hashed into a single version.

    Args:
        config_value (Union[Any, dict]): Either a single config value or a dictionary of config
            values.
    """
    if not isinstance(config_value, dict):
        return join_and_hash([str(config_value)])
    else:
        config_value = check.dict_param(config_value, "config_value")
        return join_and_hash(
            [key + resolve_config_version(val) for key, val in config_value.items()]
        )



def compute_version_for_step_output(step_context, _output):
    """
    Args:
        step_context (SystemStepExecutionContext):
        output (Output):
    """
    # Get version of solid.
    solid_def = step_context.solid_def
    # This is an incorrect placeholder implmenetation - we actually need to include the versions of all the
    # solid's inputs.
    return solid_def.version



import os
import sys
import warnings
from collections import namedtuple
from contextlib import contextmanager

import click
from click import UsageError

from dagster import check
from dagster.core.code_pointer import CodePointer
from dagster.core.definitions.reconstructable import repository_def_from_target_def
from dagster.core.host_representation import ExternalRepository, RepositoryLocation, UserProcessApi
from dagster.core.host_representation.handle import RepositoryLocationHandle
from dagster.core.instance import DagsterInstance
from dagster.core.origin import (
    PipelinePythonOrigin,
    RepositoryGrpcServerOrigin,
    RepositoryPythonOrigin,
)
from dagster.grpc.utils import get_loadable_targets
from dagster.utils.hosted_user_process import recon_repository_from_origin

from .load import (
    load_workspace_from_yaml_paths,
    location_handle_from_module_name,
    location_handle_from_python_file,
)
from .workspace import Workspace


def _cli_load_invariant(condition, msg=None):
    msg = (
        msg
        or "Invalid set of CLI arguments for loading repository/pipeline. See --help for details."
    )
    if not condition:
        raise UsageError(msg)


def _check_cli_arguments_none(kwargs, *keys):
    for key in keys:
        _cli_load_invariant(not kwargs.get(key))


def are_all_keys_empty(kwargs, keys):
    for key in keys:
        if kwargs.get(key):
            return False

    return True


WORKSPACE_CLI_ARGS = (
    "workspace",
    "python_file",
    "working_directory",
    "module_name",
    "attribute",
    "repository_yaml",
    "grpc_host",
    "grpc_port",
    "grpc_socket",
)

WorkspaceFileTarget = namedtuple("WorkspaceFileTarget", "paths")
PythonFileTarget = namedtuple("PythonFileTarget", "python_file attribute working_directory")
ModuleTarget = namedtuple("ModuleTarget", "module_name attribute")
GrpcServerTarget = namedtuple("GrpcServerTarget", "host port socket")

#  Utility target for graphql commands that do not require a workspace, e.g. downloading schema
EmptyWorkspaceTarget = namedtuple("EmptyWorkspaceTarget", "")

WorkspaceLoadTarget = (
    WorkspaceFileTarget,
    PythonFileTarget,
    ModuleTarget,
    EmptyWorkspaceTarget,
    GrpcServerTarget,
)


def created_workspace_load_target(kwargs):
    check.dict_param(kwargs, "kwargs")
    if are_all_keys_empty(kwargs, WORKSPACE_CLI_ARGS):
        if kwargs.get("empty_workspace"):
            return EmptyWorkspaceTarget()
        if os.path.exists("workspace.yaml"):
            return WorkspaceFileTarget(paths=["workspace.yaml"])
        elif os.path.exists("repository.yaml"):
            warnings.warn(
                'You are automatically loading a "repository.yaml", a deprecated '
                "capability. This capability will be eliminated in 0.9.0."
            )
            return WorkspaceFileTarget(paths=["repository.yaml"])
        raise click.UsageError("No arguments given and workspace.yaml not found.")
    if kwargs.get("repository_yaml"):
        warnings.warn(
            "You have used -y or --repository-yaml to load a workspace. "
            "This is deprecated and will be eliminated in 0.9.0."
        )
        _check_cli_arguments_none(
            kwargs,
            "python_file",
            "working_directory",
            "module_name",
            "attribute",
            "workspace",
            "grpc_host",
            "grpc_port",
            "grpc_socket",
        )
        return WorkspaceFileTarget(paths=[kwargs["repository_yaml"]])
    if kwargs.get("workspace"):
        _check_cli_arguments_none(
            kwargs,
            "python_file",
            "working_directory",
            "module_name",
            "attribute",
            "grpc_host",
            "grpc_port",
            "grpc_socket",
        )
        return WorkspaceFileTarget(paths=list(kwargs["workspace"]))
    if kwargs.get("python_file"):
        _check_cli_arguments_none(
            kwargs, "workspace", "module_name", "grpc_host", "grpc_port", "grpc_socket"
        )
        working_directory = get_working_directory_from_kwargs(kwargs)
        return PythonFileTarget(
            python_file=kwargs.get("python_file"),
            attribute=kwargs.get("attribute"),
            working_directory=working_directory,
        )
    if kwargs.get("module_name"):
        _check_cli_arguments_none(
            kwargs,
            "workspace",
            "python_file",
            "working_directory",
            "grpc_host",
            "grpc_port",
            "grpc_socket",
        )
        return ModuleTarget(
            module_name=kwargs.get("module_name"), attribute=kwargs.get("attribute"),
        )
    if kwargs.get("grpc_port"):
        _check_cli_arguments_none(kwargs, "grpc_socket")
        return GrpcServerTarget(
            port=kwargs.get("grpc_port"),
            socket=None,
            host=(kwargs.get("grpc_host") if kwargs.get("grpc_host") else "localhost"),
        )
    elif kwargs.get("grpc_socket"):
        return GrpcServerTarget(
            port=None,
            socket=kwargs.get("grpc_socket"),
            host=(kwargs.get("grpc_host") if kwargs.get("grpc_host") else "localhost"),
        )
    check.failed("invalid")


def workspace_from_load_target(load_target, instance):
    check.inst_param(load_target, "load_target", WorkspaceLoadTarget)
    check.inst_param(instance, "instance", DagsterInstance)

    opt_in_settings = instance.get_settings("opt_in")
    python_user_process_api = (
        UserProcessApi.GRPC
        if (opt_in_settings and opt_in_settings["local_servers"])
        else UserProcessApi.CLI
    )

    if isinstance(load_target, WorkspaceFileTarget):
        return load_workspace_from_yaml_paths(load_target.paths, python_user_process_api)
    elif isinstance(load_target, PythonFileTarget):
        return Workspace(
            [
                location_handle_from_python_file(
                    python_file=load_target.python_file,
                    attribute=load_target.attribute,
                    working_directory=load_target.working_directory,
                    user_process_api=python_user_process_api,
                )
            ]
        )
    elif isinstance(load_target, ModuleTarget):
        return Workspace(
            [
                location_handle_from_module_name(
                    load_target.module_name,
                    load_target.attribute,
                    user_process_api=python_user_process_api,
                )
            ]
        )
    elif isinstance(load_target, GrpcServerTarget):
        return Workspace(
            [
                RepositoryLocationHandle.create_grpc_server_location(
                    port=load_target.port, socket=load_target.socket, host=load_target.host,
                )
            ]
        )
    elif isinstance(load_target, EmptyWorkspaceTarget):
        return Workspace([])
    else:
        check.not_implemented("Unsupported: {}".format(load_target))


def get_workspace_from_kwargs(kwargs, instance):
    check.inst_param(instance, "instance", DagsterInstance)
    return workspace_from_load_target(created_workspace_load_target(kwargs), instance)


def python_target_click_options():
    return [
        click.option(
            "--python-file",
            "-f",
            type=click.Path(exists=True),
            help="Specify python file where repository or pipeline function lives.",
        ),
        click.option(
            "--module-name", "-m", help="Specify module where repository or pipeline function lives"
        ),
        click.option("--working-directory", "-d", help="Specify working directory"),
        click.option(
            "--attribute",
            "-a",
            help=(
                "Attribute that is either a 1) repository or pipeline or "
                "2) a function that returns a repository or pipeline."
            ),
        ),
    ]


def grpc_server_target_click_options():
    return [
        click.option("--grpc-port", type=click.INT, required=False),
        click.option("--grpc-socket", type=click.Path(), required=False),
        click.option("--grpc-host", type=click.STRING, required=False),
    ]


def workspace_target_click_options():
    return (
        [
            click.option("--empty-workspace", is_flag=True, help="Allow an empty workspace"),
            click.option(
                "--workspace",
                "-w",
                multiple=True,
                type=click.Path(exists=True),
                help=("Path to workspace file. Argument can be provided multiple times."),
            ),
            click.option(
                "--repository-yaml",
                "-y",
                type=click.Path(exists=True),
                help=("Path to legacy repository.yaml file"),
            ),
        ]
        + python_target_click_options()
        + grpc_server_target_click_options()
    )


def python_pipeline_target_click_options():
    return (
        python_target_click_options()
        + [
            click.option(
                "--repository",
                "-r",
                help=("Repository name, necessary if more than one repository is present."),
            )
        ]
        + [pipeline_option()]
    )


def python_pipeline_target_argument(f):
    from dagster.cli.pipeline import apply_click_params

    return apply_click_params(f, *python_pipeline_target_click_options())


def workspace_target_argument(f):
    from dagster.cli.pipeline import apply_click_params

    return apply_click_params(f, *workspace_target_click_options())


def grpc_server_origin_target_argument(f):
    from dagster.cli.pipeline import apply_click_params

    options = grpc_server_target_click_options()
    return apply_click_params(f, *options)


def python_origin_target_argument(f):
    from dagster.cli.pipeline import apply_click_params

    options = python_target_click_options()
    return apply_click_params(f, *options)


def repository_target_argument(f):
    from dagster.cli.pipeline import apply_click_params

    return apply_click_params(
        workspace_target_argument(f),
        click.option(
            "--repository",
            "-r",
            help=(
                "Repository within the workspace, necessary if more than one repository is present."
            ),
        ),
        click.option(
            "--location",
            "-l",
            help=(
                "RepositoryLocation within the workspace, necessary if more than one location is present."
            ),
        ),
    )


def pipeline_option():
    return click.option(
        "--pipeline",
        "-p",
        help=("Pipeline within the repository, necessary if more than one pipeline is present."),
    )


def pipeline_target_argument(f):
    from dagster.cli.pipeline import apply_click_params

    return apply_click_params(repository_target_argument(f), pipeline_option())


def get_repository_origin_from_kwargs(kwargs):
    load_target = created_workspace_load_target(kwargs)

    if isinstance(load_target, GrpcServerTarget):
        return RepositoryGrpcServerOrigin(
            host=load_target.host,
            port=load_target.port,
            socket=load_target.socket,
            repository_name=kwargs["repository"],
        )

    return get_repository_python_origin_from_kwargs(kwargs)


def get_pipeline_python_origin_from_kwargs(kwargs):
    repository_origin = get_repository_python_origin_from_kwargs(kwargs)
    provided_pipeline_name = kwargs.get("pipeline")

    recon_repo = recon_repository_from_origin(repository_origin)
    repo_definition = recon_repo.get_definition()

    pipeline_names = set(repo_definition.pipeline_names)

    if provided_pipeline_name is None and len(pipeline_names) == 1:
        pipeline_name = next(iter(pipeline_names))
    elif provided_pipeline_name is None:
        raise click.UsageError(
            (
                "Must provide --pipeline as there is more than one pipeline "
                "in {repository}. Options are: {pipelines}."
            ).format(repository=repo_definition.name, pipelines=_sorted_quoted(pipeline_names))
        )
    elif not provided_pipeline_name in pipeline_names:
        raise click.UsageError(
            (
                'Pipeline "{provided_pipeline_name}" not found in repository "{repository_name}". '
                "Found {found_names} instead."
            ).format(
                provided_pipeline_name=provided_pipeline_name,
                repository_name=repo_definition.name,
                found_names=_sorted_quoted(pipeline_names),
            )
        )
    else:
        pipeline_name = provided_pipeline_name

    return PipelinePythonOrigin(pipeline_name, repository_origin=repository_origin)


def _get_code_pointer_dict_from_kwargs(kwargs):
    python_file = kwargs.get("python_file")
    module_name = kwargs.get("module_name")
    working_directory = get_working_directory_from_kwargs(kwargs)
    attribute = kwargs.get("attribute")
    loadable_targets = get_loadable_targets(python_file, module_name, working_directory, attribute)
    if python_file:
        return {
            repository_def_from_target_def(
                loadable_target.target_definition
            ).name: CodePointer.from_python_file(
                python_file, loadable_target.attribute, working_directory
            )
            for loadable_target in loadable_targets
        }
    elif module_name:
        return {
            repository_def_from_target_def(
                loadable_target.target_definition
            ).name: CodePointer.from_module(module_name, loadable_target.attribute)
            for loadable_target in loadable_targets
        }
    else:
        check.failed("invalid")


def get_working_directory_from_kwargs(kwargs):
    return kwargs.get("working_directory") if kwargs.get("working_directory") else os.getcwd()


def get_repository_python_origin_from_kwargs(kwargs):
    provided_repo_name = kwargs.get("repository")

    if not (kwargs.get("python_file") or kwargs.get("module_name")):
        raise click.UsageError("Must specify a python file or module name")

    # Short-circuit the case where an attribute and no repository name is passed in,
    # giving us enough information to return an origin without loading any target
    # definitions - we may need to return an origin for a non-existent repository
    # (e.g. to log an origin ID for an error message)
    if kwargs.get("attribute") and not provided_repo_name:
        if kwargs.get("python_file"):
            code_pointer = CodePointer.from_python_file(
                kwargs.get("python_file"),
                kwargs.get("attribute"),
                get_working_directory_from_kwargs(kwargs),
            )
        elif kwargs.get("module_name"):
            code_pointer = CodePointer.from_module(
                kwargs.get("module_name"), kwargs.get("attribute"),
            )
        else:
            check.failed("Must specify a Python file or module name")
        return RepositoryPythonOrigin(executable_path=sys.executable, code_pointer=code_pointer)

    code_pointer_dict = _get_code_pointer_dict_from_kwargs(kwargs)
    if provided_repo_name is None and len(code_pointer_dict) == 1:
        code_pointer = next(iter(code_pointer_dict.values()))
    elif provided_repo_name is None:
        raise click.UsageError(
            (
                "Must provide --repository as there is more than one repository. "
                "Options are: {repos}."
            ).format(repos=_sorted_quoted(code_pointer_dict.keys()))
        )
    elif not provided_repo_name in code_pointer_dict:
        raise click.UsageError(
            'Repository "{provided_repo_name}" not found. Found {found_names} instead.'.format(
                provided_repo_name=provided_repo_name,
                found_names=_sorted_quoted(code_pointer_dict.keys()),
            )
        )
    else:
        code_pointer = code_pointer_dict[provided_repo_name]

    return RepositoryPythonOrigin(executable_path=sys.executable, code_pointer=code_pointer)


@contextmanager
def get_repository_location_from_kwargs(kwargs, instance):
    check.inst_param(instance, "instance", DagsterInstance)
    with get_workspace_from_kwargs(kwargs, instance) as workspace:
        provided_location_name = kwargs.get("location")

        if provided_location_name is None and len(workspace.repository_location_handles) == 1:
            yield RepositoryLocation.from_handle(next(iter(workspace.repository_location_handles)))

        elif provided_location_name is None:
            raise click.UsageError(
                (
                    "Must provide --location as there are more than one locations "
                    "available. Options are: {}"
                ).format(_sorted_quoted(workspace.repository_location_names))
            )

        elif not workspace.has_repository_location_handle(provided_location_name):
            raise click.UsageError(
                (
                    'Location "{provided_location_name}" not found in workspace. '
                    "Found {found_names} instead."
                ).format(
                    provided_location_name=provided_location_name,
                    found_names=_sorted_quoted(workspace.repository_location_names),
                )
            )

        else:
            yield RepositoryLocation.from_handle(
                workspace.get_repository_location_handle(provided_location_name)
            )


def get_external_repository_from_repo_location(repo_location, provided_repo_name):
    check.inst_param(repo_location, "repo_location", RepositoryLocation)
    check.opt_str_param(provided_repo_name, "provided_repo_name")

    repo_dict = repo_location.get_repositories()
    check.invariant(repo_dict, "There should be at least one repo.")

    # no name provided and there is only one repo. Automatically return
    if provided_repo_name is None and len(repo_dict) == 1:
        return next(iter(repo_dict.values()))

    if provided_repo_name is None:
        raise click.UsageError(
            (
                "Must provide --repository as there is more than one repository "
                "in {location}. Options are: {repos}."
            ).format(location=repo_location.name, repos=_sorted_quoted(repo_dict.keys()))
        )

    if not repo_location.has_repository(provided_repo_name):
        raise click.UsageError(
            (
                'Repository "{provided_repo_name}" not found in location "{location_name}". '
                "Found {found_names} instead."
            ).format(
                provided_repo_name=provided_repo_name,
                location_name=repo_location.name,
                found_names=_sorted_quoted(repo_dict.keys()),
            )
        )

    return repo_location.get_repository(provided_repo_name)


@contextmanager
def get_external_repository_from_kwargs(kwargs, instance):
    check.inst_param(instance, "instance", DagsterInstance)
    with get_repository_location_from_kwargs(kwargs, instance) as repo_location:
        provided_repo_name = kwargs.get("repository")
        yield get_external_repository_from_repo_location(repo_location, provided_repo_name)


def get_external_pipeline_from_external_repo(external_repo, provided_pipeline_name):
    check.inst_param(external_repo, "external_repo", ExternalRepository)
    check.opt_str_param(provided_pipeline_name, "provided_pipeline_name")

    external_pipelines = {ep.name: ep for ep in external_repo.get_all_external_pipelines()}

    check.invariant(external_pipelines)

    if provided_pipeline_name is None and len(external_pipelines) == 1:
        return next(iter(external_pipelines.values()))

    if provided_pipeline_name is None:
        raise click.UsageError(
            (
                "Must provide --pipeline as there is more than one pipeline "
                "in {repository}. Options are: {pipelines}."
            ).format(
                repository=external_repo.name, pipelines=_sorted_quoted(external_pipelines.keys())
            )
        )

    if not provided_pipeline_name in external_pipelines:
        raise click.UsageError(
            (
                'Pipeline "{provided_pipeline_name}" not found in repository "{repository_name}". '
                "Found {found_names} instead."
            ).format(
                provided_pipeline_name=provided_pipeline_name,
                repository_name=external_repo.name,
                found_names=_sorted_quoted(external_pipelines.keys()),
            )
        )

    return external_pipelines[provided_pipeline_name]


@contextmanager
def get_external_pipeline_from_kwargs(kwargs, instance):
    check.inst_param(instance, "instance", DagsterInstance)
    with get_external_repository_from_kwargs(kwargs, instance) as external_repo:
        provided_pipeline_name = kwargs.get("pipeline")
        yield get_external_pipeline_from_external_repo(external_repo, provided_pipeline_name)


def _sorted_quoted(strings):
    return "[" + ", ".join(["'{}'".format(s) for s in sorted(list(strings))]) + "]"



from collections import namedtuple
from enum import Enum

from dagster import check
from dagster.core.definitions import AssetMaterialization, Materialization, Solid, SolidHandle
from dagster.core.definitions.events import EventMetadataEntry
from dagster.core.types.dagster_type import DagsterType
from dagster.serdes import whitelist_for_serdes
from dagster.utils import merge_dicts
from dagster.utils.error import SerializableErrorInfo


@whitelist_for_serdes
class StepOutputHandle(namedtuple("_StepOutputHandle", "step_key output_name")):
    @staticmethod
    def from_step(step, output_name="result"):
        check.inst_param(step, "step", ExecutionStep)

        return StepOutputHandle(step.key, output_name)

    def __new__(cls, step_key, output_name="result"):
        return super(StepOutputHandle, cls).__new__(
            cls,
            step_key=check.str_param(step_key, "step_key"),
            output_name=check.str_param(output_name, "output_name"),
        )


@whitelist_for_serdes
class StepInputData(namedtuple("_StepInputData", "input_name type_check_data")):
    def __new__(cls, input_name, type_check_data):
        return super(StepInputData, cls).__new__(
            cls,
            input_name=check.str_param(input_name, "input_name"),
            type_check_data=check.opt_inst_param(type_check_data, "type_check_data", TypeCheckData),
        )


@whitelist_for_serdes
class TypeCheckData(namedtuple("_TypeCheckData", "success label description metadata_entries")):
    def __new__(cls, success, label, description=None, metadata_entries=None):
        return super(TypeCheckData, cls).__new__(
            cls,
            success=check.bool_param(success, "success"),
            label=check.str_param(label, "label"),
            description=check.opt_str_param(description, "description"),
            metadata_entries=check.opt_list_param(
                metadata_entries, metadata_entries, of_type=EventMetadataEntry
            ),
        )


@whitelist_for_serdes
class UserFailureData(namedtuple("_UserFailureData", "label description metadata_entries")):
    def __new__(cls, label, description=None, metadata_entries=None):
        return super(UserFailureData, cls).__new__(
            cls,
            label=check.str_param(label, "label"),
            description=check.opt_str_param(description, "description"),
            metadata_entries=check.opt_list_param(
                metadata_entries, metadata_entries, of_type=EventMetadataEntry
            ),
        )


@whitelist_for_serdes
class StepOutputData(
    namedtuple(
        "_StepOutputData",
        "step_output_handle intermediate_materialization type_check_data address version",
    )
):
    def __new__(
        cls,
        step_output_handle,
        intermediate_materialization=None,
        type_check_data=None,
        address=None,
        version=None,
    ):
        return super(StepOutputData, cls).__new__(
            cls,
            step_output_handle=check.inst_param(
                step_output_handle, "step_output_handle", StepOutputHandle
            ),
            intermediate_materialization=check.opt_inst_param(
                intermediate_materialization,
                "intermediate_materialization",
                (AssetMaterialization, Materialization),
            ),
            type_check_data=check.opt_inst_param(type_check_data, "type_check_data", TypeCheckData),
            address=check.opt_str_param(address, "address"),
            version=check.opt_str_param(version, "version"),
        )

    @property
    def output_name(self):
        return self.step_output_handle.output_name


@whitelist_for_serdes
class StepFailureData(namedtuple("_StepFailureData", "error user_failure_data")):
    def __new__(cls, error, user_failure_data):
        return super(StepFailureData, cls).__new__(
            cls,
            error=check.opt_inst_param(error, "error", SerializableErrorInfo),
            user_failure_data=check.opt_inst_param(
                user_failure_data, "user_failure_data", UserFailureData
            ),
        )


@whitelist_for_serdes
class StepRetryData(namedtuple("_StepRetryData", "error seconds_to_wait")):
    def __new__(cls, error, seconds_to_wait=None):
        return super(StepRetryData, cls).__new__(
            cls,
            error=check.opt_inst_param(error, "error", SerializableErrorInfo),
            seconds_to_wait=check.opt_int_param(seconds_to_wait, "seconds_to_wait"),
        )


@whitelist_for_serdes
class StepSuccessData(namedtuple("_StepSuccessData", "duration_ms")):
    def __new__(cls, duration_ms):
        return super(StepSuccessData, cls).__new__(
            cls, duration_ms=check.float_param(duration_ms, "duration_ms")
        )


@whitelist_for_serdes
class StepKind(Enum):
    COMPUTE = "COMPUTE"


class StepInputSourceType(Enum):
    SINGLE_OUTPUT = "SINGLE_OUTPUT"
    MULTIPLE_OUTPUTS = "MULTIPLE_OUTPUTS"
    CONFIG = "CONFIG"
    DEFAULT_VALUE = "DEFAULT_VALUE"


class StepInput(
    namedtuple("_StepInput", "name dagster_type source_type source_handles config_data")
):
    def __new__(
        cls, name, dagster_type=None, source_type=None, source_handles=None, config_data=None,
    ):
        return super(StepInput, cls).__new__(
            cls,
            name=check.str_param(name, "name"),
            dagster_type=check.inst_param(dagster_type, "dagster_type", DagsterType),
            source_type=check.inst_param(source_type, "source_type", StepInputSourceType),
            source_handles=check.opt_list_param(
                source_handles, "source_handles", of_type=StepOutputHandle
            ),
            config_data=config_data,  # can be any type
        )

    @property
    def is_from_output(self):
        return (
            self.source_type == StepInputSourceType.SINGLE_OUTPUT
            or self.source_type == StepInputSourceType.MULTIPLE_OUTPUTS
        )

    @property
    def is_from_single_output(self):
        return self.source_type == StepInputSourceType.SINGLE_OUTPUT

    @property
    def is_from_config(self):
        return self.source_type == StepInputSourceType.CONFIG

    @property
    def is_from_default_value(self):
        return self.source_type == StepInputSourceType.DEFAULT_VALUE

    @property
    def is_from_multiple_outputs(self):
        return self.source_type == StepInputSourceType.MULTIPLE_OUTPUTS

    @property
    def dependency_keys(self):
        return {handle.step_key for handle in self.source_handles}


class StepOutput(namedtuple("_StepOutput", "name dagster_type optional should_materialize")):
    def __new__(
        cls, name, dagster_type=None, optional=None, should_materialize=None,
    ):
        return super(StepOutput, cls).__new__(
            cls,
            name=check.str_param(name, "name"),
            optional=check.bool_param(optional, "optional"),
            should_materialize=check.bool_param(should_materialize, "should_materialize"),
            dagster_type=check.inst_param(dagster_type, "dagster_type", DagsterType),
        )


class ExecutionStep(
    namedtuple(
        "_ExecutionStep",
        (
            "pipeline_name key_suffix step_inputs step_input_dict step_outputs step_output_dict "
            "compute_fn kind solid_handle solid_version logging_tags tags hook_defs"
        ),
    )
):
    def __new__(
        cls,
        pipeline_name,
        key_suffix,
        step_inputs,
        step_outputs,
        compute_fn,
        kind,
        solid_handle,
        solid,
        logging_tags=None,
    ):
        check.inst_param(solid, "solid", Solid)

        return super(ExecutionStep, cls).__new__(
            cls,
            pipeline_name=check.str_param(pipeline_name, "pipeline_name"),
            key_suffix=check.str_param(key_suffix, "key_suffix"),
            step_inputs=check.list_param(step_inputs, "step_inputs", of_type=StepInput),
            step_input_dict={si.name: si for si in step_inputs},
            step_outputs=check.list_param(step_outputs, "step_outputs", of_type=StepOutput),
            step_output_dict={so.name: so for so in step_outputs},
            compute_fn=check.callable_param(
                compute_fn, "compute_fn"
            ),  # Compute_fn is the compute function for the step.
            #     Not to be confused with the compute_fn of the passed in solid.
            kind=check.inst_param(kind, "kind", StepKind),
            solid_handle=check.inst_param(solid_handle, "solid_handle", SolidHandle),
            solid_version=solid.definition.version,
            logging_tags=merge_dicts(
                {
                    "step_key": str(solid_handle) + "." + key_suffix,
                    "pipeline": pipeline_name,
                    "solid": solid_handle.name,
                    "solid_definition": solid.definition.name,
                },
                check.opt_dict_param(logging_tags, "logging_tags"),
            ),
            tags=solid.tags,
            hook_defs=solid.hook_defs,
        )

    @property
    def key(self):
        return str(self.solid_handle) + "." + self.key_suffix

    @property
    def solid_name(self):
        return self.solid_handle.name

    def has_step_output(self, name):
        check.str_param(name, "name")
        return name in self.step_output_dict

    def step_output_named(self, name):
        check.str_param(name, "name")
        return self.step_output_dict[name]

    def has_step_input(self, name):
        check.str_param(name, "name")
        return name in self.step_input_dict

    def step_input_named(self, name):
        check.str_param(name, "name")
        return self.step_input_dict[name]



from collections import namedtuple
from datetime import datetime

from dagster import check
from dagster.core.errors import DagsterInvalidDefinitionError
from dagster.core.instance import DagsterInstance
from dagster.core.storage.pipeline_run import PipelineRun
from dagster.core.storage.tags import check_tags
from dagster.utils import merge_dicts

from .mode import DEFAULT_MODE_NAME


class ScheduleExecutionContext(
    namedtuple("ScheduleExecutionContext", "instance scheduled_execution_time_utc")
):
    """Schedule-specific execution context.

    An instance of this class is made available as the first argument to various ScheduleDefinition
    functions. It is passed as the first argument to ``run_config_fn``, ``tags_fn``,
    and ``should_execute``.

    Attributes:
        instance (DagsterInstance): The instance configured to run the schedule
        scheduled_execution_time_utc (datetime):
            The time in UTC in which the execution was scheduled to happen. May differ slightly
            from both the actual execution time and the time at which the run config is computed.
    """

    def __new__(
        cls, instance, scheduled_execution_time_utc,
    ):

        return super(ScheduleExecutionContext, cls).__new__(
            cls,
            check.inst_param(instance, "instance", DagsterInstance),
            check.opt_inst_param(
                scheduled_execution_time_utc, "scheduled_execution_time_utc", datetime
            ),
        )


class ScheduleDefinition(object):
    """Define a schedule that targets a pipeline

    Args:
        name (str): The name of the schedule to create.
        cron_schedule (str): A valid cron string specifying when the schedule will run, e.g.,
            '45 23 * * 6' for a schedule that runs at 11:45 PM every Saturday.
        pipeline_name (str): The name of the pipeline to execute when the schedule runs.
        run_config (Optional[Dict]): The environment config that parameterizes this execution,
            as a dict.
        run_config_fn (Callable[[ScheduleExecutionContext], [Dict]]): A function that takes a
            ScheduleExecutionContext object and returns the environment configuration that
            parameterizes this execution, as a dict. You may set only one of ``run_config``
            and ``run_config_fn``.
        tags (Optional[Dict[str, str]]): A dictionary of tags (string key-value pairs) to attach
            to the scheduled runs.
        tags_fn (Optional[Callable[[ScheduleExecutionContext], Optional[Dict[str, str]]]]): A
            function that generates tags to attach to the schedules runs. Takes a
            :py:class:`~dagster.ScheduleExecutionContext` and returns a dictionary of tags (string
            key-value pairs). You may set only one of ``tags`` and ``tags_fn``.
        solid_selection (Optional[List[str]]): A list of solid subselection (including single
            solid names) to execute when the schedule runs. e.g. ``['*some_solid+', 'other_solid']``
        mode (Optional[str]): The mode to apply when executing this schedule. (default: 'default')
        should_execute (Optional[Callable[[ScheduleExecutionContext], bool]]): A function that runs
            at schedule execution time to determine whether a schedule should execute or skip. Takes
            a :py:class:`~dagster.ScheduleExecutionContext` and returns a boolean (``True`` if the
            schedule should execute). Defaults to a function that always returns ``True``.
        environment_vars (Optional[dict[str, str]]): The environment variables to set for the
            schedule
    """

    __slots__ = [
        "_name",
        "_cron_schedule",
        "_execution_params",
        "_run_config_fn",
        "_run_config",
        "_tags",
        "_mode",
        "_pipeline_name",
        "_solid_selection",
        "_tags_fn",
        "_should_execute",
        "_environment_vars",
    ]

    def __init__(
        self,
        name,
        cron_schedule,
        pipeline_name,
        run_config=None,
        run_config_fn=None,
        tags=None,
        tags_fn=None,
        solid_selection=None,
        mode="default",
        should_execute=None,
        environment_vars=None,
    ):

        self._name = check.str_param(name, "name")
        self._cron_schedule = check.str_param(cron_schedule, "cron_schedule")
        self._pipeline_name = check.str_param(pipeline_name, "pipeline_name")
        self._run_config = check.opt_dict_param(run_config, "run_config")
        self._tags = check.opt_dict_param(tags, "tags", key_type=str, value_type=str)

        check.opt_callable_param(tags_fn, "tags_fn")
        self._solid_selection = check.opt_nullable_list_param(
            solid_selection, "solid_selection", of_type=str
        )
        self._mode = check.opt_str_param(mode, "mode", DEFAULT_MODE_NAME)
        check.opt_callable_param(should_execute, "should_execute")
        self._environment_vars = check.opt_dict_param(
            environment_vars, "environment_vars", key_type=str, value_type=str
        )

        if run_config_fn and run_config:
            raise DagsterInvalidDefinitionError(
                "Attempted to provide both run_config_fn and run_config as arguments"
                " to ScheduleDefinition. Must provide only one of the two."
            )

        if not run_config and not run_config_fn:
            run_config_fn = lambda _context: {}
        self._run_config_fn = run_config_fn

        if tags_fn and tags:
            raise DagsterInvalidDefinitionError(
                "Attempted to provide both tags_fn and tags as arguments"
                " to ScheduleDefinition. Must provide only one of the two."
            )

        if not tags and not tags_fn:
            tags_fn = lambda _context: {}
        self._tags_fn = tags_fn

        if not should_execute:
            should_execute = lambda _context: True
        self._should_execute = should_execute

    @property
    def name(self):
        return self._name

    @property
    def cron_schedule(self):
        return self._cron_schedule

    @property
    def environment_vars(self):
        return self._environment_vars

    @property
    def mode(self):
        return self._mode

    def get_run_config(self, context):
        check.inst_param(context, "context", ScheduleExecutionContext)
        if self._run_config:
            return self._run_config
        return self._run_config_fn(context)

    def get_tags(self, context):
        check.inst_param(context, "context", ScheduleExecutionContext)
        if self._tags:
            tags = self._tags
            check_tags(tags, "tags")
        else:
            tags = self._tags_fn(context)
            # These tags are checked in _tags_fn_wrapper

        tags = merge_dicts(tags, PipelineRun.tags_for_schedule(self))

        return tags

    def should_execute(self, context):
        check.inst_param(context, "context", ScheduleExecutionContext)
        return self._should_execute(context)

    @property
    def pipeline_name(self):
        return self._pipeline_name

    @property
    def solid_selection(self):
        return self._solid_selection



from collections import defaultdict

import six

from dagster import check
from dagster.core.definitions import PipelineDefinition, Solid, SolidHandle
from dagster.core.definitions.events import ObjectStoreOperation
from dagster.core.definitions.solid_container import IContainSolids
from dagster.core.definitions.utils import DEFAULT_OUTPUT
from dagster.core.errors import DagsterInvariantViolationError
from dagster.core.events import DagsterEvent, DagsterEventType
from dagster.core.execution.plan.objects import StepKind


def _construct_events_by_step_key(event_list):
    events_by_step_key = defaultdict(list)
    for event in event_list:
        events_by_step_key[event.step_key].append(event)

    return dict(events_by_step_key)


class IContainSolidsExecutionResult(object):
    def __init__(self, container, event_list, reconstruct_context, handle=None):
        self.container = check.inst_param(container, "container", IContainSolids)
        self.event_list = check.list_param(event_list, "step_event_list", of_type=DagsterEvent)
        self.reconstruct_context = check.callable_param(reconstruct_context, "reconstruct_context")
        self.handle = check.opt_inst_param(handle, "handle", SolidHandle)
        self._events_by_step_key = _construct_events_by_step_key(event_list)

    @property
    def success(self):
        """bool: Whether all steps in the execution were successful."""
        return all([not event.is_failure for event in self.event_list])

    @property
    def step_event_list(self):
        """List[DagsterEvent] The full list of events generated by steps in the execution.

        Excludes events generated by the pipeline lifecycle, e.g., ``PIPELINE_START``.
        """
        return [event for event in self.event_list if event.is_step_event]

    @property
    def events_by_step_key(self):
        return self._events_by_step_key

    def result_for_solid(self, name):
        """Get the result of a top level solid.

        Args:
            name (str): The name of the top-level solid or aliased solid for which to retrieve the
                result.

        Returns:
            Union[CompositeSolidExecutionResult, SolidExecutionResult]: The result of the solid
            execution within the pipeline.
        """
        if not self.container.has_solid_named(name):
            raise DagsterInvariantViolationError(
                "Tried to get result for solid '{name}' in '{container}'. No such top level "
                "solid.".format(name=name, container=self.container.name)
            )

        return self.result_for_handle(SolidHandle(name, None))

    def output_for_solid(self, handle_str, output_name=DEFAULT_OUTPUT):
        """Get the output of a solid by its solid handle string and output name.

        Args:
            handle_str (str): The string handle for the solid.
            output_name (str): Optional. The name of the output, default to DEFAULT_OUTPUT.

        Returns:
            The output value for the handle and output_name.
        """
        check.str_param(handle_str, "handle_str")
        check.str_param(output_name, "output_name")
        return self.result_for_handle(SolidHandle.from_string(handle_str)).output_value(output_name)

    @property
    def solid_result_list(self):
        """List[Union[CompositeSolidExecutionResult, SolidExecutionResult]]: The results for each
        top level solid."""
        return [self.result_for_solid(solid.name) for solid in self.container.solids]

    def _result_for_handle(self, solid, handle):
        if not solid:
            raise DagsterInvariantViolationError(
                "Can not find solid handle {handle_str}.".format(handle_str=handle.to_string())
            )

        events_by_kind = defaultdict(list)

        if solid.is_composite:
            events = []
            for event in self.event_list:
                if event.is_step_event:
                    if event.solid_handle.is_or_descends_from(handle.with_ancestor(self.handle)):
                        events_by_kind[event.step_kind].append(event)
                        events.append(event)

            return CompositeSolidExecutionResult(
                solid,
                events,
                events_by_kind,
                self.reconstruct_context,
                handle=handle.with_ancestor(self.handle),
            )
        else:
            for event in self.event_list:
                if event.is_step_event:
                    if event.solid_handle.is_or_descends_from(handle.with_ancestor(self.handle)):
                        events_by_kind[event.step_kind].append(event)

            return SolidExecutionResult(solid, events_by_kind, self.reconstruct_context)

    def result_for_handle(self, handle):
        """Get the result of a solid by its solid handle.

        This allows indexing into top-level solids to retrieve the results of children of
        composite solids.

        Args:
            handle (Union[str,SolidHandle]): The handle for the solid.

        Returns:
            Union[CompositeSolidExecutionResult, SolidExecutionResult]: The result of the given
            solid.
        """
        if isinstance(handle, six.string_types):
            handle = SolidHandle.from_string(handle)
        else:
            check.inst_param(handle, "handle", SolidHandle)

        solid = self.container.get_solid(handle)

        return self._result_for_handle(solid, handle)


class PipelineExecutionResult(IContainSolidsExecutionResult):
    """The result of executing a pipeline.

    Returned by :py:func:`execute_pipeline`. Users should not instantiate this class.
    """

    def __init__(self, pipeline_def, run_id, event_list, reconstruct_context):
        self.run_id = check.str_param(run_id, "run_id")
        check.inst_param(pipeline_def, "pipeline_def", PipelineDefinition)

        super(PipelineExecutionResult, self).__init__(
            container=pipeline_def, event_list=event_list, reconstruct_context=reconstruct_context,
        )

    @property
    def pipeline_def(self):
        return self.container


class CompositeSolidExecutionResult(IContainSolidsExecutionResult):
    """Execution result for a composite solid in a pipeline.

    Users should not instantiate this class.
    """

    def __init__(self, solid, event_list, step_events_by_kind, reconstruct_context, handle=None):
        check.inst_param(solid, "solid", Solid)
        check.invariant(
            solid.is_composite,
            desc="Tried to instantiate a CompositeSolidExecutionResult with a noncomposite solid",
        )
        self.solid = solid
        self.step_events_by_kind = check.dict_param(
            step_events_by_kind, "step_events_by_kind", key_type=StepKind, value_type=list
        )

        super(CompositeSolidExecutionResult, self).__init__(
            container=solid.definition,
            event_list=event_list,
            reconstruct_context=reconstruct_context,
            handle=handle,
        )

    def output_values_for_solid(self, name):
        check.str_param(name, "name")
        return self.result_for_solid(name).output_values

    def output_values_for_handle(self, handle_str):
        check.str_param(handle_str, "handle_str")

        return self.result_for_handle(handle_str).output_values

    def output_value_for_solid(self, name, output_name=DEFAULT_OUTPUT):
        check.str_param(name, "name")
        check.str_param(output_name, "output_name")

        return self.result_for_solid(name).output_value(output_name)

    def output_value_for_handle(self, handle_str, output_name=DEFAULT_OUTPUT):
        check.str_param(handle_str, "handle_str")
        check.str_param(output_name, "output_name")

        return self.result_for_handle(handle_str).output_value(output_name)

    @property
    def output_values(self):
        values = {}

        for output_name in self.solid.definition.output_dict:
            output_mapping = self.solid.definition.get_output_mapping(output_name)

            inner_solid_values = self._result_for_handle(
                self.solid.definition.solid_named(output_mapping.solid_name),
                SolidHandle(output_mapping.solid_name, None),
            ).output_values

            if inner_solid_values is not None:  # may be None if inner solid was skipped
                if output_mapping.output_name in inner_solid_values:
                    values[output_name] = inner_solid_values[output_mapping.output_name]

        return values

    def output_value(self, output_name=DEFAULT_OUTPUT):
        check.str_param(output_name, "output_name")

        if not self.solid.definition.has_output(output_name):
            raise DagsterInvariantViolationError(
                "Output '{output_name}' not defined in composite solid '{solid}': "
                "{outputs_clause}. If you were expecting this output to be present, you may "
                "be missing an output_mapping from an inner solid to its enclosing composite "
                "solid.".format(
                    output_name=output_name,
                    solid=self.solid.name,
                    outputs_clause="found outputs {output_names}".format(
                        output_names=str(list(self.solid.definition.output_dict.keys()))
                    )
                    if self.solid.definition.output_dict
                    else "no output mappings were defined",
                )
            )

        output_mapping = self.solid.definition.get_output_mapping(output_name)

        return self._result_for_handle(
            self.solid.definition.solid_named(output_mapping.solid_name),
            SolidHandle(output_mapping.solid_name, None),
        ).output_value(output_mapping.output_name)


class SolidExecutionResult(object):
    """Execution result for a leaf solid in a pipeline.

    Users should not instantiate this class.
    """

    def __init__(self, solid, step_events_by_kind, reconstruct_context):
        check.inst_param(solid, "solid", Solid)
        check.invariant(
            not solid.is_composite,
            desc="Tried to instantiate a SolidExecutionResult with a composite solid",
        )
        self.solid = solid
        self.step_events_by_kind = check.dict_param(
            step_events_by_kind, "step_events_by_kind", key_type=StepKind, value_type=list
        )
        self.reconstruct_context = check.callable_param(reconstruct_context, "reconstruct_context")

    @property
    def compute_input_event_dict(self):
        """Dict[str, DagsterEvent]: All events of type ``STEP_INPUT``, keyed by input name."""
        return {se.event_specific_data.input_name: se for se in self.input_events_during_compute}

    @property
    def input_events_during_compute(self):
        """List[DagsterEvent]: All events of type ``STEP_INPUT``."""
        return self._compute_steps_of_type(DagsterEventType.STEP_INPUT)

    @property
    def compute_output_event_dict(self):
        """Dict[str, DagsterEvent]: All events of type ``STEP_OUTPUT``, keyed by output name"""
        return {se.event_specific_data.output_name: se for se in self.output_events_during_compute}

    def get_output_event_for_compute(self, output_name="result"):
        """The ``STEP_OUTPUT`` event for the given output name.

        Throws if not present.

        Args:
            output_name (Optional[str]): The name of the output. (default: 'result')

        Returns:
            DagsterEvent: The corresponding event.
        """
        return self.compute_output_event_dict[output_name]

    @property
    def output_events_during_compute(self):
        """List[DagsterEvent]: All events of type ``STEP_OUTPUT``."""
        return self._compute_steps_of_type(DagsterEventType.STEP_OUTPUT)

    @property
    def compute_step_events(self):
        """List[DagsterEvent]: All events generated by execution of the solid compute function."""
        return self.step_events_by_kind.get(StepKind.COMPUTE, [])

    @property
    def step_events(self):
        return self.compute_step_events

    @property
    def materializations_during_compute(self):
        """List[Materialization]: All materializations yielded by the solid."""
        return [
            mat_event.event_specific_data.materialization
            for mat_event in self.materialization_events_during_compute
        ]

    @property
    def materialization_events_during_compute(self):
        """List[DagsterEvent]: All events of type ``STEP_MATERIALIZATION``."""
        return self._compute_steps_of_type(DagsterEventType.STEP_MATERIALIZATION)

    @property
    def expectation_events_during_compute(self):
        """List[DagsterEvent]: All events of type ``STEP_EXPECTATION_RESULT``."""
        return self._compute_steps_of_type(DagsterEventType.STEP_EXPECTATION_RESULT)

    def _compute_steps_of_type(self, dagster_event_type):
        return list(
            filter(lambda se: se.event_type == dagster_event_type, self.compute_step_events)
        )

    @property
    def expectation_results_during_compute(self):
        """List[ExpectationResult]: All expectation results yielded by the solid"""
        return [
            expt_event.event_specific_data.expectation_result
            for expt_event in self.expectation_events_during_compute
        ]

    def get_step_success_event(self):
        """DagsterEvent: The ``STEP_SUCCESS`` event, throws if not present."""
        for step_event in self.compute_step_events:
            if step_event.event_type == DagsterEventType.STEP_SUCCESS:
                return step_event

        check.failed("Step success not found for solid {}".format(self.solid.name))

    @property
    def compute_step_failure_event(self):
        """DagsterEvent: The ``STEP_FAILURE`` event, throws if it did not fail."""
        if self.success:
            raise DagsterInvariantViolationError(
                "Cannot call compute_step_failure_event if successful"
            )

        step_failure_events = self._compute_steps_of_type(DagsterEventType.STEP_FAILURE)
        check.invariant(len(step_failure_events) == 1)
        return step_failure_events[0]

    @property
    def success(self):
        """bool: Whether solid execution was successful."""
        any_success = False
        for step_event in self.compute_step_events:
            if step_event.event_type == DagsterEventType.STEP_FAILURE:
                return False
            if step_event.event_type == DagsterEventType.STEP_SUCCESS:
                any_success = True

        return any_success

    @property
    def skipped(self):
        """bool: Whether solid execution was skipped."""
        return all(
            [
                step_event.event_type == DagsterEventType.STEP_SKIPPED
                for step_event in self.compute_step_events
            ]
        )

    @property
    def output_values(self):
        """Union[None, Dict[str, Any]]: The computed output values.

        Keys of this dictionary are output names, values are output values.

        Returns ``None`` if execution did not succeed.

        Note that accessing this property will reconstruct the pipeline context (including, e.g.,
        resources) to retrieve materialized output values.
        """
        from .api import create_execution_plan

        if self.success and self.compute_step_events:
            with self.reconstruct_context() as context:
                execution_plan = create_execution_plan(
                    pipeline=context.pipeline,
                    run_config=context.run_config,
                    mode=context.pipeline_run.mode,
                    step_keys_to_execute=context.pipeline_run.step_keys_to_execute,
                )
                values = {}
                for compute_step_event in self.compute_step_events:
                    if compute_step_event.is_successful_output:
                        values[compute_step_event.step_output_data.output_name] = self._get_value(
                            context.for_step(
                                execution_plan.get_step_by_key(compute_step_event.step_key)
                            ),
                            compute_step_event.step_output_data,
                        )

                return values
        else:
            return None

    def output_value(self, output_name=DEFAULT_OUTPUT):
        """Get a computed output value.

        Note that calling this method will reconstruct the pipeline context (including, e.g.,
        resources) to retrieve materialized output values.

        Args:
            output_name(str): The output name for which to retrieve the value. (default: 'result')

        Returns:
            Union[None, Any]: ``None`` if execution did not succeed, otherwise the output value.
        """
        from .api import create_execution_plan

        check.str_param(output_name, "output_name")

        if not self.solid.definition.has_output(output_name):
            raise DagsterInvariantViolationError(
                "Output '{output_name}' not defined in solid '{solid}': found outputs "
                "{output_names}".format(
                    output_name=output_name,
                    solid=self.solid.name,
                    output_names=str(list(self.solid.definition.output_dict.keys())),
                )
            )

        if self.success:
            for compute_step_event in self.compute_step_events:
                if (
                    compute_step_event.is_successful_output
                    and compute_step_event.step_output_data.output_name == output_name
                ):
                    with self.reconstruct_context() as context:
                        execution_plan = create_execution_plan(
                            pipeline=context.pipeline,
                            run_config=context.run_config,
                            mode=context.pipeline_run.mode,
                            step_keys_to_execute=context.pipeline_run.step_keys_to_execute,
                        )

                        value = self._get_value(
                            context.for_step(
                                execution_plan.get_step_by_key(compute_step_event.step_key)
                            ),
                            compute_step_event.step_output_data,
                        )
                    return value

            raise DagsterInvariantViolationError(
                (
                    "Did not find result {output_name} in solid {self.solid.name} "
                    "execution result"
                ).format(output_name=output_name, self=self)
            )
        else:
            return None

    def _get_value(self, context, step_output_data):
        value = context.intermediate_storage.get_intermediate(
            context=context,
            dagster_type=self.solid.output_def_named(step_output_data.output_name).dagster_type,
            step_output_handle=step_output_data.step_output_handle,
        )
        if isinstance(value, ObjectStoreOperation):
            return value.obj

        return value

    @property
    def failure_data(self):
        """Union[None, StepFailureData]: Any data corresponding to this step's failure, if it
        failed."""
        for step_event in self.compute_step_events:
            if step_event.event_type == DagsterEventType.STEP_FAILURE:
                return step_event.step_failure_data
